/* draft mbed TextLCD 
 * (c) 2007/8, sford
 */
 
#include "TextLCD_I2C.h"

#include "mbed.h"


/*
 * useful info found at http://www.a-netz.de/lcd.en.php
 *
 *
 * Initialisation
 * ==============
 *
 * After attaching the supply voltage/after a reset, the display needs to be brought in to a defined state
 *
 * - wait approximately 15 ms so the display is ready to execute commands
 * - Execute the command 0x30 ("Display Settings") three times (wait 1,64ms after each command, the busy flag cannot be queried now). 
 * - The display is in 8 bit mode, so if you have only connected 4 data pins you should only transmit the higher nibble of each command.
 * - If you want to use the 4 bit mode, now you can execute the command to switch over to this mode now.
 * - Execute the "clear display" command
 *
 * Timing
 * ======
 *
 * Nearly all commands transmitted to the display need 40us for execution. 
 * Exceptions are the commands "Clear Display and Reset" and "Set Cursor to Start Position" 
 * These commands need 1.64ms for execution. These timings are valid for all displays working with an 
 * internal clock of 250kHz. But I do not know any displays that use other frequencies. Any time you 
 * can use the busy flag to test if the display is ready to accept the next command.
 * 
 * _e is kept high apart from calling clock
 * _rw is kept 0 (write) apart from actions that uyse it differently
 * _rs is set by the data/command writes


  * Chris additionals
  
  RS = 0
  RW = 1
  E  = 2
  D4 = 4
  D5 = 5  
  D6 = 6
  D7 = 7  
  
 */

TextLCD_I2C::TextLCD_I2C(PinName sda, PinName scl, int addr) : _i2c(sda,scl) {

    _rows = 2;
    _columns = 16;

    _iodata = 0x0;
        
    _addr = addr; // mask in the instance address

    _rw(0);
    _e(1);
    _rs(0); // command mode
    

    // Should theoretically wait 15ms, but most things will be powered up pre-reset
    // so i'll disable that for the minute. If implemented, could wait 15ms post reset
    // instead
    // wait(0.015); 
        
    // send "Display Settings" 3 times (Only top nibble of 0x30 as we've got 4-bit bus)
    for(int i=0; i<3; i++) {
        writeNibble(0x3);
        wait(0.00164);      // this command takes 1.64ms, so wait for it
    }
    writeNibble(0x2); // 4-bit mode
            
    writeCommand(0x28);    // Function set 001 BW N F - -  
    writeCommand(0x0C);
    writeCommand(0x6);  //  Cursor Direction and Display Shift : 0000 01 CD S (CD 0-left, 1-right S(hift) 0-no, 1-yes
    
    cls();
}

int TextLCD_I2C::_putc(int value) {
    if(value == '\n') {
        newline();
    } else {
        writeData(value);
    }
    return value;
}

int TextLCD_I2C::_getc() {
    return 0;
}

void TextLCD_I2C::newline() {
    _column = 0;
    _row++;
    if(_row >= _rows) {
        _row = 0;
    }
    locate(_column, _row); 
}

void TextLCD_I2C::locate(int column, int row) {
    if(column < 0 || column >= _columns || row < 0 || row >= _rows) {
        // error("locate(%d,%d) out of range on %dx%d display", column, row, _columns, _rows);
        return;
    }
    
    _row = row;
      _column = column;
      int address = 0x80 + (_row * 40) + _column; // memory starts at 0x80, and is 40 chars long per row
      writeCommand(address);            
}

void TextLCD_I2C::cls() {
    writeCommand(0x01); // Clear Display
    wait(0.00164f);        // This command takes 1.64 ms
      locate(0, 0);
}

void TextLCD_I2C::reset() {
    cls();
}

void TextLCD_I2C::clock() {
    wait(0.000040f);
    _e(0);
    wait(0.000040f);  // most instructions take 40us
    _e(1);    
}

void TextLCD_I2C::writeNibble(int value) {
    _d(value);
    clock();
}

void TextLCD_I2C::writeByte(int value) {
    writeNibble(value >> 4);
    writeNibble(value >> 0);
}

void TextLCD_I2C::writeCommand(int command) {
    _rs(0);
    writeByte(command);
}

void TextLCD_I2C::writeData(int data) {
    _rs(1);
    writeByte(data);
    _column++;
    if(_column >= _columns) {
        newline();
    } 
}

void TextLCD_I2C::led(int data) {
    // set _iodata[4] then write the byte
    if(data==0) {
        _iodata &= 0x07f;} // force to 0         
    else {
        _iodata |= 0x80;} // force to 1    
    i2cwrite(_iodata);
}

void TextLCD_I2C::_rs(int data) {
    // set _iodata[4] then write the byte
    if(data==0) {
        _iodata &= 0xef;} // force to 0         
    else {
        _iodata |= 0x10;} // force to 1    
    i2cwrite(_iodata);
}


void TextLCD_I2C::_rw(int data) {
    // set _iodata[5] then write the byte
    if(data==0) {
        _iodata &= 0xdf;} // force to 0         
    else {
        _iodata |= 0x20;} // force to 1    
    i2cwrite(_iodata);
}

void TextLCD_I2C::_e(int data) {
    // set _iodata[6] then write the byte
    if(data==0) {
        _iodata &= 0xbf;} // force to 0         
    else {
        _iodata |= 0x40;} // force to 1    
    i2cwrite(_iodata);
}

void TextLCD_I2C::_d(int data) {

// shift lowest nibble to highest nibble, and write
_iodata &= 0xf0; // clear lower nibble
_iodata |= (data & 0xf);
i2cwrite(_iodata);
}



void TextLCD_I2C::i2cwrite(int data) {
    char foo[1];
    foo[0] = data;
    _i2c.write(_addr,foo,1);
    return;
 }


